/*
 *  Licensed Materials - Property of IBM
 *  
 *  CICS SupportPac CA1T
 * 
 *   Copyright IBM Corporation 2008, 2013
 *  All Rights Reserved
 *  
 *  US Government Users Restricted Rights  Use,
 *  duplication, or disclosure restricted by GSA ADP
 *  Schedule Contract with IBM Corporation
 */

/*
 *      File Name:    HAExit.java
 *      Version:      CA1T V2.2 
 *      Date:         07012013
 *      Product:      CICS Transaction Gateway
 *      
 *      This abstract class defines the common functions used by the FailOver and RoundRobin classes
 *      as well as providing the request validation logic
 *      
 *      See SupportPac documentation for further details on how to configure and use this sample
 *      
 * Change History
 * V1.0   - First version 
 * V1.1   - Comment parsing
 * V2.0   - Retry interval and V8 LastServers support
 * V2.0.1 - No request validation for XA, and call type logging 
 * V2.1   - Add support for an alternate policy
 * V2.2   - Add support for inclusive vs exclusive request validation   
 * 
 */

package com.ibm.ctg.samples.ha;

import com.ibm.ctg.ha.CICSRequestExit;
import com.ibm.ctg.ha.ExitEvent;
import com.ibm.ctg.ha.ExitEventData;
import com.ibm.ctg.ha.InvalidRequestException;
import com.ibm.ctg.ha.RequestDetails;
import com.ibm.ctg.ha.RequestPayloadType;
import com.ibm.ctg.ha.RequestCallType;
import com.ibm.ctg.samples.ha.HAUtil;

import java.util.*;

public abstract class HAExit implements CICSRequestExit {
   //                                             
   public static final String copyright = "\nLicensed Materials - Property of IBM "+
         "CICS SupportPac CA1T" +
         "(c) Copyright IBM Corp. 2008, 2013  All Rights Reserved. " +
         "US Government Users Restricted Rights - Use, duplication or " +
         "disclosure restricted by GSA ADP Schedule Contract with " +
         "IBM Corp.\n";
   //


   protected int refresh; // Time for stale servers to expire
   protected HAUtil util = new HAUtil(); //Instance of helper class
   protected static String mainConfFile;// Config file name
   protected static String altConfFile;// Alternative config file name
   protected boolean enableFlag = true; // Disable/enable exit

   //Stateful objects used by HA logic
   protected Hashtable<String, String[]> mainTable = null;
   protected Hashtable<String, String[]> altTable = null;
   protected Hashtable<String, String[]> configTable ; // Parsed config file
   protected Hashtable<String, Date> dirtyTable ; // Table of unavailable servers
   protected Hashtable<String, Integer> serverTable ; // Table of servers for round robin policy
   protected int  nextIndex = 0; // Index in roundrobin table

   // Request validation objects
   protected boolean reqval;       // Flag to indicate if reqest validation is required
   protected boolean newreqval;   // New value of reqval
   protected Set<String> transidSet;   // Set of tranids for validation logic
   protected Set<String> useridSet;  // Set of users for validation logic
   protected Hashtable<String, String> progTable; // Program and payload type table for validation logic
   private long autoSwapTimeout = 0;
   private Calendar swapTime=null;
   private boolean isUsableCfg = false;  // Guilty until proven innocent
   private boolean isInclusive = true;  // Validation rules default to inclusive

   /**
    * Constructor called at exit initialisation by Gateway 
    */
   public HAExit() throws Exception {

      try {
         util.initTraceLevel();    // set tracing level
         String destStr=util.initLog();        // Setup output destination
         String validationStyle="INCLUSIVE";   // Set validation style with default
         util.println("CICS request exit %s V%s",getClass().getSimpleName(),HAUtil.VERSION);
         util.println("Logging destination: %s",destStr);
         refresh = util.getRetryInterval(); // Set retry interval
         autoSwapTimeout = util.getAutoSwap(); // Set auto swap timeout
         mainConfFile = util.getMainConfigFile(); // Get name of config file
         altConfFile = util.getAltConfigFile(); // Get alternative config
         isInclusive = util.getValidationStyle(); // Get validation style

         initConfig(); // Load config files and build tables

         if (!isInclusive) { 
            validationStyle="EXCLUSIVE";
         }

         if (util.isAltConfig()) {
            util.println("Configuration parameters: Intervals(retry=%ds autoswap=%ds) Logging level(%d) Validation style(%s)",
                  refresh / 1000, autoSwapTimeout / 1000, util.getTraceLevel(), validationStyle);
         } else {
            util.println("Configuration parameters: Retry Interval(%ds) Logging level(%d) Validation style(%s)",
                  refresh / 1000, util.getTraceLevel(), validationStyle);
         }
         if (!isUsableCfg) {
            util.println("The main policy contains no valid rules. The CICS request exit is inactive.");
         }

      } catch (Exception e) {
         util.println ("Terminating with exception %s",e.toString());
         throw new Exception (e);
      }
   }

   /*
    * 
    * Builds sets for use during program validation
    * 
    * Input: String denoting key type and table of config file values
    * Output: Set of values
    */
   public Set<String> buildSet (String type, Hashtable<String,String[]> table) {

      Set<String> newSet = new HashSet<String>(); // temp set

      if (table.get(type) != null) {
         newreqval = true; // Set reqval flag to true if we found a rule
         String[] array = table.get(type); 

         for ( int i=0;i < array.length;i++) {
            newSet.add(array[i]);
         } 
      } 
      return newSet;
   }

   /*
    * 
    * Builds tables for use during program validation
    * 
    * Input: String denoting key type and Table of config file values
    * Output: Table of String=String for PROGRAM=payloadtype
    */
   public Hashtable<String, String> buildTable (String type, Hashtable<String,String[]> table) {

      Hashtable<String, String> newTable = new Hashtable<String, String>(); // temp table

      if (table.get(type) != null) {
         newreqval = true; // Set reqval flag to true if we found a rule
         String [] array = table.get(HAUtil.PROGID);
         // Check if prog values specify payload type

         for (int i=0; i < array.length;i++) {
            String program = array[i];
            util.debugln("Raw program value(%s)",program);
            String payloadType = "NULL";  // Defaults is any type of payload
            if (program.contains("/")) {               
               payloadType = program.substring(program.indexOf("/") + 1);
               program = program.substring(0, program.indexOf("/"));
            }
            newTable.put(program, payloadType);
            util.debugln("Adding program(%s) payload(%s)",program,payloadType); 
         } 
      }   
      return newTable;
   }

   /*
    * Parses CREXIT commands and sets trace levels and reload events
    * 
    * Input: ExitEvent and Command Map
    * Output: void
    */
   public synchronized void eventFired(ExitEvent event, Map<ExitEventData,Object> data) {

      // Following command strings are valid
      String level0 = HAUtil.id.trim()+"_0";
      String level1 = HAUtil.id.trim()+"_1";
      String level2 = HAUtil.id.trim()+"_2";
      String level3 = HAUtil.id.trim()+"_3";
      String reload = HAUtil.id.trim()+"_RELOAD";
      String reset  = HAUtil.id.trim()+"_RESET";
      String list  = HAUtil.id.trim()+"_LIST";
      String file  = HAUtil.id.trim()+"_FILE";
      String disable  = HAUtil.id.trim()+"_DISABLE";
      String enable  = HAUtil.id.trim()+"_ENABLE";
      String swap = HAUtil.id.trim()+"_SWAP";

      try {
         if (event.equals(ExitEvent.ShutDown)) {
            shutdown();
         } 
         else if (event.equals(ExitEvent.Command)) {

            String command = ((String) data.get(ExitEventData.CommandData)).trim();
            util.println( "Command [%s] received",command);

            if (command.equalsIgnoreCase(level0)) {
               util.setTraceLevel(0);
               util.infoln ("Output disabled");
            }
            else if (command.equalsIgnoreCase(level1)) {
               util.setTraceLevel(1);
               util.infoln ("Level 1 tracing enabled");
            } 
            else if (command.equalsIgnoreCase(level2)) {
               util.setTraceLevel(2);
               util.infoln ("Level 2 tracing enabled");
            }
            else if (command.equalsIgnoreCase(level3)) {
               util.setTraceLevel(3);
               util.infoln ("Level 3 tracing enabled");
            }
            else if (command.equalsIgnoreCase(reload)) {
               util.infoln ("Configuration reload initiated");
               initConfig();
               if (!isUsableCfg) {
                  util.println("The main policy contains no valid rules. The CICS request exit is inactive.");
               }

            }
            else if (command.equalsIgnoreCase(reset)) {
               if (enableFlag) {
                  util.infoln ("CICS server state reset initiated");
                  setState();
               } else {
                  util.infoln ("CICS server state ignored - exit is disabled");
               }
            }
            else if (command.equalsIgnoreCase(list)) {
               util.infoln ("Configuration list initiated");
               listState();
            }
            else if (command.equalsIgnoreCase(disable)) {
               util.infoln ("Disabled CICS request exit");
               enableFlag=false;
            }
            else if (command.equalsIgnoreCase(enable)) {
               util.infoln ("Enabled CICS request exit");
               enableFlag=true;
               setState();
            } else if (command.equalsIgnoreCase(swap)) {
               util.infoln("Swapping configuration files");
               swapConfig();
            }
            else if (command.startsWith(file)) {   
               if (command.contains("=")) {
                  String userCfgSpec = command.substring(command.indexOf("=")+1);
                  util.setConfigSpec(userCfgSpec);
                  util.infoln ("Configuration spec updated");
                  mainConfFile = util.getMainConfigFile(); // Get name of config file
                  altConfFile = util.getAltConfigFile(); // Get alternative config
                  initConfig();
               } 
               else {
                  util.println ("Error: invalid file command");
               }
            }
            else { 
               util.println("Error: invalid command ");
            }
         }         

         // Unknown event fired
         else {
            util.println("Unknown event fired: %s",event.toString() );
         }
      }
      catch (Exception e) {
         e.printStackTrace();
      }
   }

   /**
    * Reads configuration files
    * Synchronised along with mapServer
    * Input: Name of configuration file
    * Output: Re-populates configTable
    */
   public synchronized void initConfig() throws Exception {

      try {
         // Read the main configuration file
         util.println("Loading main policy file %s", mainConfFile);
         mainTable = util.readfile(mainConfFile); // Read configuration file
         String configLabel=String.format("Policy(%s)",HAUtil.MAIN_POLICY);


         // We have attempted to load defined policy files
         // Complete initialization only if MAIN policy is non-empty
         if (mainTable.isEmpty()) {
            isUsableCfg = false;
            util.println("Main policy file %s contained no recognized entries.",mainConfFile);
            isUsableCfg = false;
         } else {
            isUsableCfg=true;
            configTable = mainTable; // All OK so update configuration table
            setState(); // Reset server status based on the new configTable
            listConfig(configLabel,mainTable);
         }

         // We appear to have a valid main policy, look for alternative policy
         if (util.isAltConfig()) {
            // Read the alternative configuration file
            util.println("Loading alternate policy file %s", altConfFile);
            altTable = util.readfile(altConfFile);
            if (altTable.isEmpty()) {
               util.println("Alternate policy file %s contained no recognized entries. It is ignored.",
                     altConfFile);
               altTable = null;
            } else {
               configLabel=String.format("Policy(%s)",HAUtil.ALT_POLICY);
               listConfig(configLabel,altTable);
            }
         }
         listState();

      } catch (Exception e) {
         util.println("File load failed with exception %s", e);
         throw new Exception(e);
      }
   }


   /**
    * Reads configuration files 
    * Synchronised along with mapServer 
    */
   public synchronized void swapConfig() throws Exception {

      if (isUsableCfg) {
         if (altTable != null) {
            if (isMainActive()) {
               util.infoln("Swapping to alternative policy: %s", altConfFile);
               configTable = altTable;
            } else {
               util.infoln("Swapping to main policy %s", mainConfFile);
               configTable = mainTable;
            }
            // Store time of swap
            swapTime = Calendar.getInstance();

            setState(); // Reset server status based on the new configTable
            listState(); // Log server state
         } else {
            // no alternative to swap too!
            util.infoln("No alternative policy available. Swap request is ignored.");
         }
      } else {
         util.infoln("Main policy file contains no recognized entries. Swap ignored.");
      }
   }

   /**
    * Reads config files 
    * Synchronised along with mapServer 
    */
   public synchronized void autoSwap() throws Exception {

      if (isUsableCfg) { 
         // Swap back to main config after given time
         if ((autoSwapTimeout>0) && configTable.equals(altTable)) {
            Calendar now=Calendar.getInstance();
            long timeSince=now.getTimeInMillis()-swapTime.getTimeInMillis();
            util.debugln("autoSwap() : %dms since alternative config selected. autoSwapTimeout=%d",timeSince,autoSwapTimeout);
            if (timeSince>autoSwapTimeout) {
               util.infoln("Autoswap timeout has expired. Swapping back to main policy %s",mainConfFile);
               swapConfig();
            }
         } else {
            // else nothing to do - just debug trace
            util.debugln("autoSwap() - Nothing to do. Active policy is main %s",mainConfFile);
         }
      } else {
         util.debugln("autoSwap() - Exit is inactive as the main policy is unusable");
      }
   }

   /**
    * List status 
    * Output: Prints list of server rules & states
    */
   public synchronized void listState() throws Exception {
      String configLabel=String.format("Active policy(%s)",HAUtil.MAIN_POLICY);
      // Print active policy if alternative is available
      if (util.isAltConfig()) {
         if (isMainActive()) {
            //            util.println("%s from %s",configLabel,mainConfFile);
         } else {
            configLabel=String.format("Active policy(%s)",HAUtil.ALT_POLICY);
            //            util.println("%s from %s",configLabel,altConfFile);
         }
      }

      // Print out list of server rules
      printServerRules(configLabel,configTable);

      // Print out dirty table state 
      printDirtyServers(dirtyTable);

      // Print out server validation rules
      printServerValidationRules(configLabel,configTable);
   }

   /**
    * List config 
    * Output: Prints list of server rules
    */
   public synchronized void listConfig(String configLabel, Hashtable<String, String[]> aCfgTable) throws Exception {
      // Print out list of server rules
      printServerRules(configLabel,aCfgTable);

      // Print out server validation rules
      printServerValidationRules(configLabel,aCfgTable);
   }

   /**
    * List server based rules
    * Output: Prints list of CICS server RULES
    */
   void printServerRules(String configLabel, Hashtable<String, String[]> aCfgTable) {
      // Print out list of server rules
      if (aCfgTable != null) {
         String[] stra = new String[aCfgTable.values().size()];// max array size

         Iterator<String> i = aCfgTable.keySet().iterator();
         while (i.hasNext()) {
            String key = i.next().toString(); // server key value from config
            // file
            if (!key.equalsIgnoreCase(HAUtil.PROGID)
                  && !key.equalsIgnoreCase(HAUtil.TRANSID)
                  && !key.equalsIgnoreCase(HAUtil.USERID)) {

               stra = (aCfgTable.get(key)); // Array of configured servers
               util.println("%s: Server rule %s using actual servers %s",
                     configLabel, key,Arrays.toString(stra));
            }
         }
      }
   }

   /**
    * List validation rules
    * Output: Prints CICS server validation rules
    */
   void printServerValidationRules(String configLabel, Hashtable<String, String[]> aCfgTable) {
      // Print validation rules
      String[] stra=null;
      stra = aCfgTable.get(HAUtil.PROGID);
      if (stra!=null) util.println("%s: PROGRAM validation rules set for %s",
            configLabel,Arrays.toString(stra));

      // Print validation rules
      stra = aCfgTable.get(HAUtil.TRANSID);
      if (stra!=null) util.println("Policy %s: TRANSID validation rules set for %s",
            configLabel,Arrays.toString(stra));

      // Print validation rules
      stra = aCfgTable.get(HAUtil.USERID);
      if (stra!=null) util.println("Policy %s: USERID validation rules set for %s",
            configLabel,Arrays.toString(stra));
   }

   /**
    * List dirty servers
    * Output: Prints table of dirty CICS servers
    */
   void printDirtyServers(Hashtable<String, Date> aServerStateTable) {
      // Print out list of dirty servers
      if ( (aServerStateTable!=null) && 
            !aServerStateTable.isEmpty() ) {     
         String[] stra=null;
         int size = aServerStateTable.size(); // size of dirty table
         util.debugln("printDirtyServers() - dirty server table has %d servers",size);

         // First spin around all servers in dirty table and check if any have expired
         stra = new String[size];//max array size
         stra = aServerStateTable.keySet().toArray(stra);
         Date now = new Date();
         util.debugln("printDirtyServers() - Checking if servers in dirty table have expired");
         for (int n =0; n < size; n++) {
            isClean(stra[n],now); //Delete any server if expired  
         }

         // Log any servers left in dirty table
         size = aServerStateTable.size(); // new size of dirty table
         if (size > 0) {
            stra = new String[size];//max array size
            util.debugln("printDirtyServers() - Dirty server table now has %d servers",size);
            stra = aServerStateTable.keySet().toArray(stra); //  Get all servers in the dirty table
            util.println("Following servers are currently unavailable %s",Arrays.toString(stra));
         }
         else {
            util.println("All CICS servers are currently clean");
         }
      } else {
         util.debugln("printDirtyServers() - nothing to print");
      }

   }

   /**
    * Initialises all tables
    * Synchronised along with mapServer
    */
   public synchronized void setState() throws Exception {

      // Temporary tables for use during this method
      Hashtable<String, Integer>  newserverTable = new Hashtable<String,Integer>() ;
      Hashtable<String, Date>     newdirtyTable = new Hashtable<String,Date>() ;

      Set<String> newtransidSet;
      Set<String> newuseridSet;
      Hashtable<String, String> newprogTable;

      try {         
         util.debugln("Initialising CICS server table");
         Iterator<String> i = configTable.keySet().iterator();
         while (i.hasNext()){
            String key = i.next().toString();
            if (!key.equalsIgnoreCase(HAUtil.PROGID)
                  && !key.equalsIgnoreCase(HAUtil.TRANSID)
                  && !key.equalsIgnoreCase(HAUtil.USERID)) {
               newserverTable.put(key, 0);
            }
         }
         util.debugln("Built server rules for %s",newserverTable.keySet().toString());

         // Build request validation rules      
         newreqval=false;   
         newtransidSet = buildSet(HAUtil.TRANSID,configTable);
         if (!newtransidSet.isEmpty()) {
            util.debugln("Built %s rules for %s",HAUtil.TRANSID,newtransidSet.toString());
         }

         newuseridSet  = buildSet(HAUtil.USERID,configTable);
         if (!newuseridSet.isEmpty()) {
            util.debugln("Built %s rules for %s",HAUtil.USERID,newuseridSet.toString());
         }

         newprogTable  = buildTable(HAUtil.PROGID,configTable);
         if (!newprogTable.isEmpty()) {
            util.debugln("Built %s rules for %s",HAUtil.PROGID,newprogTable.keySet().toString());
         }

      } catch (Exception e) {
         util.println("Initialisation failed with exception %s",e.toString());
         throw new Exception(e);
      }

      // All OK so set update the state tables and reqval 
      transidSet=newtransidSet;
      useridSet=newuseridSet;
      progTable=newprogTable;
      dirtyTable=newdirtyTable;
      serverTable=newserverTable;
      reqval=newreqval;

   }

   /**
    * getCICSServer is called by Gateway for each request. 
    * This is the main method in the exit. 
    * It first checks the server mapping rules then invokes the mapServer method
    * to validate the request details
    * Input: RequestDetails map
    * Output: server name
    */
   public String getCICSServer(Map<RequestDetails, Object> data) throws InvalidRequestException {

      Object dataItem;
      String serveralias; // input server from ECI request
      String cicsserver = null; // Output server
      Object calltype;
      int retry = (Integer) data.get(RequestDetails.RetryCount);
      boolean remap = false;
      boolean xa = false;
      String thread = Thread.currentThread().getName(); // ID of current worker thread
      Date d = new Date(); 

      // Autoswap will switch policies if the timer has popped
      try {
         autoSwap();
      } catch (Exception e) {
         util.println("Autoswap failed with %s",e.toString());
         // Attempt to continue with existing config, whatever it is
      }

      // Set the call type
      dataItem = data.get(RequestDetails.RequestType);
      // set XA flag so we can disable validation and payload rules
      util.debugln ("getCICSServer() - Request type: %s",dataItem.toString());
      if (dataItem.equals(RequestCallType.XA))
      {
         xa = true;
         util.debugln ("getCICSServer() - XA request detected");   
      }
      calltype = dataItem.toString();

      // Set the server alias for the mapping rules
      dataItem = data.get(RequestDetails.Server);

      // Short circuit if we are inactive with no policy
      if (!isUsableCfg) {
         String quickPick=null;
         if (dataItem!=null) {
            quickPick=dataItem.toString();
         }
         util.debugln("Thread(%s) Server name remapping disabled. Return original request alias %s",thread,quickPick);
         return quickPick;
      }

      if (dataItem != null) {
         serveralias = dataItem.toString(); 
      } else if (enableFlag) {
         serveralias = HAUtil.NULL_SERVER; // if alias is null use NULL rule if remapping enabled
      } else {
         serveralias = null; 
      }


      // First validate any requests
      // reqval flag is set if we have validation rules
      // enableFlag is false if exit disabled
      if ((reqval)&&(enableFlag)&&(!xa)) {
         util.debugln ("getCICSServer() - Validating request");
         validateReq(data, serveralias, calltype);
      } else {
         util.debugln ("getCICSServer() - Validation disabled");         
      }

      if (enableFlag) {
         // If retrying build error message and mark last server used as dead
         if (retry > 0){
            String lastserver = data.get(RequestDetails.LastServer).toString();
            String error = data.get(RequestDetails.LastError).toString();
            setDirty(lastserver,error,d); // Add last server to the dirty table 
         }

         try {
            cicsserver = remapServer(data, serveralias, calltype, xa, d);
            remap = true;
         } catch (NoCICSServersException ncse1) {
            // is there an alternative policy ?
            if (util.isAltConfig()) {
               // yes - Are we currently using the main policy?
               if (isMainActive()) {
                  // Yes - ok, try to remap after switching to alternate
                  util.debugln("NoCICSServersException: %s",ncse1.getMessage());
                  util.traceln("Main config servers are all dirty, attempt autoswap to alternate config");
                  try {
                     swapConfig();
                     // retry the mapping with alternate config
                     cicsserver = remapServer(data, serveralias, calltype, xa, d);
                     remap = true;
                  } catch (NoCICSServersException ncse2) {
                     // Handle NoCICSServersException exception from remapServer()                     
                     util.debugln("NoCICSServersException: %s",ncse2.getMessage());
                     util.traceln("Alternate policy servers are all dirty, no choices left");
                     // no servers available in the alternate config either
                     cicsserver=getOldestDirty();
                     if (cicsserver==null) {
                        throw new InvalidRequestException(HAUtil.id + "Error selecting eldest dirty CICS server");
                     }
                     remap=true;
                  } catch (Exception e) {
                     // General exception from swapConfig() - fail the request
                     util.println("Main config servers are all dirty, "
                           + "and autoswap to alternate config failed with "
                           + e);
                     throw new InvalidRequestException(HAUtil.id + "General exception",e);
                  }
               } else {
                  // Last chance - return eldest dirty server
                  util.traceln("Thread(%s) Call(%s) Retry(%d) server alias(%s) request rejected. All servers currently dirty. ",
                        thread,calltype.toString(),retry,serveralias);
                  cicsserver=getOldestDirty();
                  if (cicsserver==null) {
                     // end of the line - give up!
                     throw new InvalidRequestException(HAUtil.id + "Error selecting eldest dirty CICS server. The request is rejected.");
                  }
                  remap=true;
               }
            } else {
               util.traceln("Main policy servers are all dirty. No alternate policy available.");
            }
         }         
      } else {
         // Exit is disabled so return original alias
         cicsserver = serveralias;
         util.debugln("Thread(%s) Server name remapping inactive as there is no usable Main policy. Return original request alias %s",thread,serveralias);
         if (retry>=1) {
            int lastError = (Integer) data.get(RequestDetails.LastError);
            throw new InvalidRequestException(HAUtil.id + " request retry disabled, last error("+lastError+")");
         }
      }

      // If server name remapped log a message
      if (remap) {
         util.traceln("Thread(%s) Call(%s) Retry(%d) remapped server alias(%s) to actual server(%s)",
               thread,calltype.toString(),retry,serveralias,cicsserver);
      }
      else {
         util.traceln("Thread(%s) Call(%s) Retry(%d) no server rule matched for server alias(%s)",
               thread,calltype.toString(),retry,serveralias);
      }

      return cicsserver;  //Return to the CTG with the new CICS server name
   }


   /**
    * Invoked by Gateway at startup to set the max number of retries
    * 
    * Output: retry count
    */

   public int getRetryCount() {
      Integer count;

      count = util.getMaxRetry(); // Get value of environment variable
      if (count == -1) {
         util.debugln("getRetryCount() - %s not set, invoking setRetryCount()",HAUtil.ENV_MAXRETRY);
         count = setRetryCount(); // Get retry count from server array lengths
      }

      if (count < 2) {
         // IF we have only one or zero servers defined, retry count must be zero
         count=0;
      } else {
         // ELSE it is one less than number of servers
         count-=1;
      }
      util.infoln("Set maximum retry count (%d)",count);
      return count;  
   }
   /**
    * Dynamically calculate max retry count based on longser server array rule
    * Output: integer value of max array length
    */
   public int setRetryCount() {
      int retry=0;

      if (isUsableCfg) {
         retry += util.getRuleMaxLen(mainTable);
         util.debugln("setRetryCount() main policy longest rule=%d", retry);
         if (altTable != null) {
            int altMaxLen = util.getRuleMaxLen(altTable);
            retry += altMaxLen;
            util.debugln("setRetryCount() alternate policy longest rule=%d",
                  altMaxLen);
         }
         util.traceln("setRetryCount() dynamically set to %d", retry);
      } else {
         // default retry to zero when no policy
         util.traceln("setRetryCount() no usable main policy; set to zero");
      }

      return retry;
   }

   /**
    * 
    * Checks if server in the dirty table is past the retry interval
    * Removes dirty servers from dirty table if current time is past retry interval
    * 
    * Inputs: Server to be checked, current time 
    * Returns: true if server is ok to be used
    */
   public boolean isClean(String server, Date now) {

      boolean flag = false;
      Date lastused = dirtyTable.get(server); // last time server was used
      if (lastused != null) { // If server found in table check if retry interval has expired
         long time = now.getTime() - lastused.getTime();
         if (time >= refresh) {
            flag = true;
            setClean(server);
         } else{
            util.debugln("isClean() CICS server (%s) will remain in the dirty table for %dms",server,(refresh-time));
         }
      }else { // Server was not in table so no worries
         flag = true;
         util.debugln("isClean() CICS server (%s) not in dirty table",server);
      }
      return flag;
   }

   /**
    * Maps a server alias to a server
    * Abstract class implemented by FailOver and RoundRobin classes  
    */
   public abstract String mapServer(String alias, Date d, Map<RequestDetails, Object> data) throws NoCICSServersException;



   /**
    * Returns name of the current class
    */
   public String myName() {
      return this.getClass().getCanonicalName();
   }

   /**
    * Removes dirty servers from dirty table
    * Input: Server
    */
   public void setClean(String server) {
      util.traceln("Server(%s) is now eligible for requests",server);
      dirtyTable.remove(server);
   }

   /**
    * Adds dirty server to dirty table with time stamp
    * Input: server, last error, date
    */
   public void setDirty(String server, String error, Date d) {
      util.traceln("Setting last server(%s) dirty - Error(%s)",server,error);
      dirtyTable.put(server, d);
   }

   /**
    * Called at Shutdwon of the Gateway
    */
   public void shutdown() {

      util.println("Shutdown initiated for CICS request exit %s",getClass().getSimpleName());
   }

   /**
    * Performs request validation 
    */
   public void validateReq(Map<RequestDetails, Object> data, String server, Object calltype)
         throws InvalidRequestException {


      String program, transid, userid, payload;
      Object dataObject;
      String thread = Thread.currentThread().getName(); // ID of current worker thread

      // Validate TRANSACTION
      dataObject = data.get(RequestDetails.Transid);
      if (dataObject == null) {
         transid = "NULL";
      } else {
         transid = dataObject.toString();
      }      
      // check if the request parameters are not specified in the set
      if (transidSet.isEmpty()) {
         util.debugln("Bypassed transid validation");         
      } else {
         boolean filterMatchTransaction = false;
         filterMatchTransaction = transidSet.contains(transid.toUpperCase());

         if ((isInclusive && !filterMatchTransaction) ||
               (!isInclusive && filterMatchTransaction)) {
            util.infoln("Rejected transid(%s) to server(%s)",transid,server);
            throw new InvalidRequestException(HAUtil.id + " Rejected(" + transid+ ")");
         }
         util.debugln("Validated transid(%s) to server(%s)",transid,server);
      }

      // Validate PROGRAM and/or PAYLOAD TYPE for program
      dataObject = data.get(RequestDetails.Program);
      if (dataObject == null) {
         program = "NULL";
      } else {
         program = dataObject.toString();
      }

      dataObject = data.get(RequestDetails.PayloadType);
      if (dataObject == null) {
         payload = "NULL";
      } else {
         payload = dataObject.toString();
      }

      // check if the Channel Commarea type is correct
      if ( progTable.isEmpty() ) {
         util.debugln("Bypassed program/payload validation");
      } else if (!progTable.containsKey(program.toUpperCase())) {
         // is there any program entry in the table?
         util.debugln("No validation rule for program(%s)",program);
      } else if ( progTable.get(program.toUpperCase()).equalsIgnoreCase("NULL") ) {
         // there is but is there no payload rule?
         util.debugln("Validation rule for program(%s), no payload specified",program);
         boolean filterMatchProgram = false;
         filterMatchProgram = progTable.containsKey(program.toUpperCase());
         if ((isInclusive && !filterMatchProgram) ||
               (!isInclusive && filterMatchProgram)) {
            util.infoln(HAUtil.id + "Rejected program(%s) to server(%s)",program,server);
            throw new InvalidRequestException(HAUtil.id + " rejected(" + program+ ")");
         }
      } else {
         // there is a program rule and there is a payload rule
         boolean filterMatchPayload = false;
         filterMatchPayload = progTable.get(program.toUpperCase()).equalsIgnoreCase(payload);
         util.debugln("Program(%s) request payload(%s) vs. rule payload(%s) matched=%b",
               program,payload,progTable.get(program.toUpperCase()),filterMatchPayload);
         if ((isInclusive && !filterMatchPayload) ||
               (!isInclusive && filterMatchPayload)) {
            util.infoln(HAUtil.id + "Rejected program(%s) payload(%s) to server(%s)",program,payload,server);
            throw new InvalidRequestException(HAUtil.id + "rejected program("+program+") type("+payload+")" );
         } 
         util.debugln("Validated payload(%s) for program(%s)",payload,program);
      }

      // Validate USERID 
      dataObject = data.get(RequestDetails.Userid);
      if (dataObject == null) {
         userid = "NULL";
      } else {
         userid = dataObject.toString();
      }
      // Check if user is specified in the userid list
      if (useridSet.isEmpty()){
         util.debugln("Bypassed userid validation");
      } else {
         boolean filterMatchUserid = false;
         filterMatchUserid= useridSet.contains(userid.toUpperCase());      
         if ((isInclusive && !filterMatchUserid) ||
               (!isInclusive && filterMatchUserid)){
            util.infoln("Rejected userid(%s) to server(%s)",userid,server);
            throw new InvalidRequestException(HAUtil.id + " rejected(" + userid+ ")");
         } 
         util.debugln("Validated userid(%s) to server(%s)",userid,server);         
      }

      // All OK    
      util.traceln("Thread(%s) Call(%s) Validated request server(%s) transid(%s) userid(%s) program(%s) payload(%s)",
            thread,
            calltype.toString(),
            server,
            transid,
            userid,
            program,
            payload); 
   }

   protected synchronized boolean isMainActive() {
      return configTable.equals(mainTable);
   }

   protected synchronized boolean isAltActive() {
      return configTable.equals(altTable);
   }

   protected synchronized String remapServer(Map<RequestDetails, Object> data,
         String serveralias, Object calltype, boolean xa, Date d) throws NoCICSServersException {
      String cicsserver = null; // Output server
      int retry = (Integer) data.get(RequestDetails.RetryCount);
      String thread = Thread.currentThread().getName(); // ID of current worker thread
      Object dataItem;
      Object payload = null;

      // Set the payload type as commarea or channel
      dataItem = data.get(RequestDetails.PayloadType);         
      if (dataItem != null) {
         payload = dataItem; 
         util.debugln ("remapServer() - set payload %s",payload );  
      } else  {
         payload = "NULL"; // 
         util.debugln ("remapServer() - No payload detected");   
      }

      // Check if we have a specific rule for this alias and remap the request
      if (configTable.containsKey(serveralias)) {
         util.traceln("Thread(%s) Call(%s) Retry(%d) server alias(%s) matched server rule(%s)",
               thread,
               calltype.toString(),
               retry,
               serveralias,
               serveralias);
         cicsserver = mapServer(serveralias, d, data);

         // Check if ipic rule for channels and not XA request
      } else if ((!xa)
            && (configTable.containsKey(HAUtil.IPIC_SERVER))
            && (payload.equals(RequestPayloadType.CHANNEL))) {
         util.traceln("Thread(%s) Call(%s) Retry(%d) server alias(%s) matched server rule(%s) with payload(%s)",
               thread,
               calltype.toString(),
               retry,
               serveralias,
               HAUtil.IPIC_SERVER,
               RequestPayloadType.CHANNEL.toString());

         cicsserver = mapServer(HAUtil.IPIC_SERVER, d, data);

         // Check if exci rule for commareas and not XA request
      } else if ((!xa)
            && (configTable.containsKey(HAUtil.EXCI_SERVER))
            && (payload.equals(RequestPayloadType.COMMAREA))) {
         util.traceln("Thread(%s) Call(%s) Retry(%d) server alias(%s) matched server rule(%s) with payload(%s)",
               thread,
               calltype.toString(),
               retry,
               serveralias,
               HAUtil.EXCI_SERVER,
               RequestPayloadType.COMMAREA.toString());
         cicsserver = mapServer(HAUtil.EXCI_SERVER, d, data);  

         // Server rule not specified so use * server rule
      } else if (configTable.containsKey(HAUtil.ANY_SERVER)) {
         util.traceln("Thread(%s) Call(%s) Retry(%d) server alias(%s) matched server rule(%s)",
               thread,
               calltype.toString(),
               retry,
               serveralias,
               HAUtil.ANY_SERVER);
         cicsserver = mapServer(HAUtil.ANY_SERVER, d, data);

         // No mapping so return original alias
      } else {
         util.debugln("Thread(%s) Call(%s) Retry(%d) server alias(%s) no rule matched server(%s)",
               thread,
               calltype.toString(),
               serveralias);
         cicsserver = serveralias;
      }
      return cicsserver;
   }

   protected synchronized String getOldestDirty() {
      String oldestDirty=null;
      Date eldestDate=null;
      Date[] dirtyDates={};
      if ((dirtyTable!=null) && (!dirtyTable.isEmpty())) {
         // Find eldest entry
         Collection<Date> cd=dirtyTable.values();
         util.debugln("Set of %d date values in the dirty table %s",cd.size(),cd.toString());
         dirtyDates=dirtyTable.values().toArray(dirtyDates);
         if (dirtyDates==null) {
            util.debugln("getOldestDirty() dirty server array is null");
         }
         util.debugln("getOldestDirty() dirty server date set is %s",dirtyDates.toString());
         Arrays.sort(dirtyDates);
         util.debugln("getOldestDirty() sorted dirty server date set is %s",dirtyDates.toString());
         eldestDate=dirtyDates[0];
         util.debugln("getOldestDirty() eldest dirty server is %s",eldestDate.toString());
         // Locate server corresponding to eldest entry
         Iterator<String> keys=dirtyTable.keySet().iterator();
         while(keys.hasNext() && oldestDirty==null) {
            String server=keys.next();
            if (dirtyTable.get(server).equals(eldestDate)) {
               oldestDirty=server;
            }
         }
         util.traceln("getOldestDirty() eldest dirty server is %s from %s",oldestDirty,eldestDate.toString());
      } else {
         util.traceln("getOldestDirty() no dirty servers - something went wrong!");
      }
      return oldestDirty;
   }

}