/*
 *  Licensed Materials - Property of IBM
 *  
 *  CICS SupportPac CA1T
 * 
 *   Copyright IBM Corporation 2008, 2013
 *  All Rights Reserved
 *  
 *  US Government Users Restricted Rights  Use,
 *  duplication, or disclosure restricted by GSA ADP
 *  Schedule Contract with IBM Corporation
 */

/*
 *      File Name:    HAUtil.java
 *      Version:      CA1T V2.2 
 *      Date:         07012013
 *      Product:      CICS Transaction Gateway
 * 
 *      Utility class for CICS Transaction Gatway SupportPac CA1T
 *      Provides constant values and control of output functions
 *      
 *      
 * Change History
 * V1.0 - GA version
 * V1.1 - Comment parsing
 * V2.0 - Retry interval and V8 LastServers support
 * V2.1 - Add support for an alternate policy   
 * V2.2 - Add support for inclusive vs exclusive request validation   
 * 
 */

package com.ibm.ctg.samples.ha;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

//ZFileRecordReader.newReader is deprecated in JZOS 2.4.2
//Replace ZFile and ZFileRecordReader with the following 
//when JZOS 2.4.2 is the min level supported by CA1T
//import com.ibm.jzos.RecordReader;
//import com.ibm.jzos.ZFileConstants;
import com.ibm.jzos.ZFile;
import com.ibm.jzos.ZFileRecordReader;
import com.ibm.jzos.FileFactory;


public class HAUtil {
   //                                             
   public static final String copyright = "\nLicensed Materials - Property of IBM "+
         "CICS SupportPac CA1T" +
         "(c) Copyright IBM Corp. 2008, 2013  All Rights Reserved. " +
         "US Government Users Restricted Rights - Use, duplication or " +
         "disclosure restricted by GSA ADP Schedule Contract with " +
         "IBM Corp.\n";
   //

   PrintStream stdout; // Print stream for output

   public static final String VERSION = "3.0";
   static final String id = " CA1T "; // SupportPac ID
   static final String EBCDIC = "IBM037"; // Dataset encoding for reading file

   static final String ENV_FILE = "CTG_HACONFIG"; // Config file environment variable
   static final String ENV_INTERVAL = "CTG_HAINTERVAL"; // Interval environment variable
   static final String ENV_MAXRETRY = "CTG_HACOUNT"; // Max retry environment variable
   static final String ENV_AUTOSWAP = "CTG_HAAUTOSWAP"; // Autoswap time environment variable
   static final String ENV_VALIDATION = "CTG_HAVALIDATION"; // Validation style environment variable

   static final int REFRESH = 60000; // Default retry interval (ms) for stale servers
   static final int AUTOSWAP = 60000; // Default interval (ms) to autoswap alt config

   static final String ANY_SERVER = "*"; // Any server rule
   static final String NULL_SERVER = "NULL"; // null server rule
   static final String EXCI_SERVER = "COMMAREA"; // commarea server rule
   static final String IPIC_SERVER = "CHANNEL"; // channel server rule

   //Statics to set the ini file key values for request parsing
   public static final String TRANSID = "TRANSID"; 
   public static final String USERID = "USERID";
   public static final String PROGID = "PROGRAM";

   static final String FILESPEC_SEPARATOR=";";
   static final String comma = ","; // index delimiter in config file
   static final String delim = "="; // key delimiter in config file

   static final String STDOUT_PROPERTY = "com.ibm.ctg.samples.haexit.out"; // logging destination
   static final String DEBUGLEVEL_PROPERTY = "com.ibm.ctg.samples.haexit.ca1t"; // debug level

   static final String dateFmt = "MM/dd/yy HH:mm:ss:SSS"; // Format of time stamps

   static final String MAIN_POLICY = "Main";
   static final String ALT_POLICY = "Alternative";

   private String mainConfigFile=null;
   private String altConfigFile=null;
   private static int traceLevel = 1; // Default tracing level

   /**
    * Setup the stdout logging destination for message
    * Default is Gateway STDOUT, but can be set to DD
    */
   public String initLog()
   {
      // set up output destination
      String destination = System.getProperty(STDOUT_PROPERTY);
      String stdoutDescription=null;

      if (destination == null) {
         stdout = getPrintStream();
         stdoutDescription = "default (stdout)";
      } else {

         File destFile = new File(destination);

         // Use com.ibm.jzos.FileFactory to handle DD output
         try {
            if (destination.substring(0, 2).equalsIgnoreCase("//")) {
               BufferedOutputStream bufOut = FileFactory.newBufferedOutputStream(destination);
               stdout = new PrintStream(bufOut);
               stdoutDescription = String.format("MVS file: %s",destination);
            } else {
               stdout = new PrintStream(destFile);
               stdoutDescription = String.format("file: %s",destination);
            }

         } catch (FileNotFoundException e) {
            stdout = getPrintStream();
            System.err.printf("%s Print to '%s' failed with %s\n",id,destination,e.toString());
            System.err.printf("%s Using default PrintStream\n",id);
         } catch (IOException e) {
            stdout = getPrintStream();
            System.err.printf("%s Print to '%s' failed with %s\n",id,destination,e.toString());
            System.err.printf("%s Using default PrintStream\n",id);
         }
      }
      return stdoutDescription;
   }

   /**
    * Reads environment variable ENV_FILE
    * Sets global main and alt config file names
    */
   private void parseConfigFileNames(String userCfgSpec) throws Exception {

      if (userCfgSpec== null) {
         throw new IOException(ENV_FILE + " variable not set");
      } else {
         debugln("parseConfigFileNames(%s)",userCfgSpec);
      }
      try {
         String fileList[]=userCfgSpec.split(FILESPEC_SEPARATOR);
         if (fileList.length>2) {
            throw new IOException(ENV_FILE + " contains too many entries. The maximum is 2.");             
         } else {
            // split will at least return the contents of configfile unmodified in fileList[0]
            // or the first filename in the list
            mainConfigFile=fileList[0];
            debugln("parseConfigFileNames() - Set configuration file: %s",mainConfigFile);
            if (fileList.length>1) {
               // we have an alternate config file
               altConfigFile=fileList[1];
               debugln("parseConfigFileNames() - Set alternate configuration file: %s",altConfigFile);
            }
         }

      } catch (IOException ioe) {
         println ("Error: " + ioe.toString());
         throw new Exception (ioe);
      }
   }
   /**
    * Reads environment variable specified by ENV_MAXRETRY
    * Returns numeric value of environment variable  
    */
   public int getMaxRetry() {
      int retry = -1; // default to -1 variable not set

      String s = System.getenv(ENV_MAXRETRY); // Read variable
      debugln("getMaxRetry() - %s=%s",ENV_MAXRETRY,s);

      if (s!=null ){

         try {
            int i = new Integer(s); // convert string to integer value
            if (i>=1) {
               retry = i;
            } else {
               debugln("getMaxRetry() - invalid value for %s=%s",ENV_MAXRETRY,s);
               retry =1;
            }

         }
         catch (NumberFormatException e) {
            infoln("Error: NumberFormatException parsing %s value (%s)",ENV_MAXRETRY,s);
         }
      }
      return retry;

   }

   /**
    * Reads environment variable specified by ENV_INTERVAL
    * Returns retry interval in ms 
    */
   public int getRetryInterval() {
      int interval;
      Integer i;
      String s;

      s = System.getenv(ENV_INTERVAL);
      debugln("getRefreshInterval() - %s=%s",ENV_INTERVAL,s);

      if (s == null) {
         interval = REFRESH;
      }else {
         i = new Integer(s);
         interval = (int)(i *1000);
      }

      return interval;
   }

   /**
    * Reads environment variable specified by ENV_AUTOSWAP
    * Returns retry interval in ms 
    */
   public int getAutoSwap() {
      int autoswap;
      Integer i;
      String s;

      s = System.getenv(ENV_AUTOSWAP);
      debugln("getAutoSwap() - %s=%s",ENV_AUTOSWAP,s);

      if (s == null) {
         // return default value
         autoswap = AUTOSWAP;
      }else {
         i = new Integer(s);
         autoswap = (int)(i *1000);
      }

      return autoswap;
   }

   /**
    * Reads environment variable specified by ENV_VALIDATION
    * Returns retry interval in ms 
    */
   public boolean getValidationStyle() {
      boolean inclusive=true;
      String s;

      s = System.getenv(ENV_VALIDATION);
      debugln("getValidationStyle() - %s=%s",ENV_VALIDATION,s);

      if (s != null) {
         if (s.equalsIgnoreCase("EXCLUSIVE")) {
            inclusive=false;
         }
      }  // else return default value

      return inclusive;
   }   

   /**
    * Returns PrintStream to use as default output logging devic
    */
   public PrintStream getPrintStream() {
      return System.out;
   }

   /**
    * Send a string to current PrintWriter
    */
   public void print( String args) {

      try {
         stdout.print(args);

      } catch (Throwable t) {
         // We cannot output anything so throw exception to disable exit
         throw new RuntimeException("Default PrintStream has failed!");
      }
   }

   /**
    * Get set of values from a line of text
    * Input: Line of text
    * Returns String array of values
    */
   public String[] readvalues (String line, String key) throws Exception  {

      String value; // Values from config file
      String array[] = null; // Array of values
      debugln ("readvalues() - Line: %s key: %s",line,key);

      // Parse the line to get the key location          
      int keyEnd = line.indexOf(delim); // find KEY
      debugln ("readvalues() - keyend: %d",keyEnd);
      if ( keyEnd >= 1 ) {
         value = line.substring(keyEnd + 1);
      }
      else {
         debugln ("Unable to locate key:%s in line: %s",key,line);
         throw new Exception ("Error parsing file");
      }
      value = line.substring(keyEnd+1,line.length()).toUpperCase(); // find VALUE
      debugln("readvalues() - key = %s value = %s",key,value);

      // Build VALUE into array of tokens
      StringTokenizer tokenizer = new StringTokenizer(value,comma);
      int count = tokenizer.countTokens();  // count the number of values 
      array = new String[count]; //initialise array   

      int c = 0;
      while (tokenizer.hasMoreTokens()) {
         array[c] = tokenizer.nextToken();
         debugln("readvalues() - Added value %s to array",array[c].toString()); 
         c++;
      }      
      return(array); // return array of values
   }


   /**
    * Get key from a line of text
    * Input: Line of text
    * Output: Key
    */

   public String readkey (String line) throws Exception  {

      String key = null; // key from config file       
      // Parse the line to read the key          
      int keyEnd = line.indexOf(delim); // find KEY
      if ((keyEnd > 0 ) && (keyEnd < line.length())) {
         key = line.substring(0,keyEnd).toUpperCase();
      } else {
         debugln ("readkey() - keyEnd=%d line length=%d",keyEnd,line.length());
         throw new IOException("Key parsing failed for line: " + line);
      }
      debugln("readkey() - located key: %s",key);
      return key;
   }


   /**
    * Read and parse configuration file
    * Lines are of the format KEY=VALUE,VALUE,VALUE
    * 
    * Returns new Hash table, containing an array of values in each row
    */

   public Hashtable<String, String[]> readfile (String file) throws Exception {

      String strLine; // Line to be read
      Hashtable<String, String[]> newTable = new Hashtable<String, String[]>(); // Table

      try {
         // Read MVS File with jzos
         if (file.substring(0, 2).equalsIgnoreCase("//")) {
            debugln("Reading MVS dataset %s",file);
            // ZFileRecordReader.newReader() is deprecated in JZOS 2.4.2
            // but Java 6.0.0 (TG V8.0) shipped JZOS 2.4.0.
            // Java 6.0.1 (TG V8.1) shipped 2.4.2.
            ZFile zfile = new ZFile(file, "rb,type=record,noseek");
            byte[] recordBuf = new byte[zfile.getLrecl()];
            @SuppressWarnings("deprecation")
            ZFileRecordReader reader = ZFileRecordReader.newReader(zfile);
            // Use the following with JZOS 2.4.2 when it is the min
            //             RecordReader reader = RecordReader.newReader(file, ZFileConstants.FLAG_DISP_SHR);
            //             byte[] recordBuf = new byte[reader.getLrecl()];
            // loop through file until no more bytes read
            while (reader.read(recordBuf) >= 0) {
               strLine = new String(recordBuf,EBCDIC);      
               strLine = strLine.trim(); 
               if ((strLine.length() > 0) && !strLine.startsWith("#")) { 
                  String key=readkey(strLine); // Get key from this line
                  newTable.put(key, readvalues(strLine,key)); // Add key and values to table
               }
            }
            reader.close();
         } 

         // Reading a HFS file
         else 
         {
            debugln("Reading HFS configuration file %s",file);
            FileReader frProperties = new FileReader(file);
            LineNumberReader lnrProperties = new LineNumberReader(frProperties);

            // loop through file until null line is read
            while (((strLine = lnrProperties.readLine()) != null)) {
               strLine = strLine.trim();// Trim  whitespace
               if ((strLine.length() > 0) && !strLine.startsWith("#")) {               
                  String key = readkey(strLine); // Get key from this line
                  newTable.put(key, readvalues(strLine,key)); // Add key and values to table
               }
            }
         }

      } catch (FileNotFoundException e) {
         throw e;
      } catch (IOException e) {
         throw e;
      } catch (Exception e) {
         e.printStackTrace();

      }
      return newTable;
   }

   /* 
    * check system property to determine trace level
    */
   public void initTraceLevel() {

      String debug = System.getProperty(DEBUGLEVEL_PROPERTY);
      if (debug != null ) {
         if (debug.equalsIgnoreCase("0")){
            traceLevel=0;    }
         else if (debug.equalsIgnoreCase("1")){
            traceLevel=1;             
         }
         else if (debug.equalsIgnoreCase("2")){
            traceLevel=2;
         }
         else if (debug.equalsIgnoreCase("3")){
            traceLevel=3;
         }
      }
   }

   /* 
    * Return the current trace level
    */
   public int getTraceLevel() {
      return traceLevel;
   }

   /* 
    * set trace level
    */
   public void setTraceLevel(int newLevel) {
      // ignore if outside of range
      int oldLevel=traceLevel;
      if ((newLevel > -1) && (newLevel < 4)) {
         traceLevel=newLevel;
      }
      debugln("setTraceLevel(newLevel=%d) oldLevel=%d traceLevel=%d",newLevel,oldLevel,traceLevel);
   }

   /**
    * Send a string to the PrintWriter and flush
    * Will output trace lines regardless of trace level
    */
   public void printf( String string, Object ... objs) {

      if ( string != null ) {
         Date timestamp = new Date();
         SimpleDateFormat dfTime = null;
         dfTime = (SimpleDateFormat) DateFormat.getTimeInstance();
         dfTime.applyPattern(dateFmt);
         String time = dfTime.format(timestamp);

         String body=String.format(string,objs);
         String message=String.format("%s%s%s",time,id,body);
         print(message);
         if (stdout == null) {
            System.out.flush();
         } else {
            stdout.flush();
         }
      }
   }

   /**
    * Send a string to the PrintWriter, append a newline, and flush
    * Will output trace lines regardless of trace level
    */
   public void println( String string, Object ... objs) {

      if ( string != null ) {
         Date timestamp = new Date();
         SimpleDateFormat dfTime = null;
         dfTime = (SimpleDateFormat) DateFormat.getTimeInstance();
         dfTime.applyPattern(dateFmt);
         String time = dfTime.format(timestamp);

         String body=String.format(string,objs);
         String message=String.format("%s%s%s\n",time,id,body);
         print(message);
         if (stdout == null) {
            System.out.flush();
         } else {
            stdout.flush();
         }
      }
   }

   /**
    * Output routine to trace a string to our currently set PrintWriter
    * Will output trace lines if trace level 1 is set
    */
   public void infoln(String string, Object ... objs) {
      if ((string != null) && (traceLevel>=1)) {
         println(string,objs);
      }
   }

   /**
    * Output routine to trace a string to our currently set PrintWriter
    * Will output trace lines if trace level 2 is set
    */
   public void traceln(String string, Object ... objs) {
      if ((string != null) && (traceLevel>=2)) {
         println(string,objs);
      }
   }

   /**
    * Output routine to trace a string to our currently set PrintWriter
    * Will output trace lines if trace level 3 is enabled
    */
   public void debugln(String string, Object ... objs) {
      if ((string != null) && (traceLevel==3)) {
         println(string,objs);
      }
   }

   /**
    * Test if an alternative configuration file is defined
    */
   public boolean isAltConfig() {
      boolean altConfig=false;
      if (altConfigFile!=null) {
         altConfig=true;
      }
      return altConfig;
   }

   public String getMainConfigFile() throws Exception {
      // Check for mainConfigFile already set. 
      // If this is already set no need to parse env var. 
      if (mainConfigFile==null) {
         String userCfgSpec=System.getenv(ENV_FILE);
         parseConfigFileNames(userCfgSpec);
      }
      return mainConfigFile;
   }

   public String getAltConfigFile() throws Exception {
      // Check for mainConfigFile already set. 
      // If this is already set no need to parse env var.
      // Check is purposely against main since it is the min.
      if (mainConfigFile==null) {
         String userCfgSpec=System.getenv(ENV_FILE);
         parseConfigFileNames(userCfgSpec);
      }

      return altConfigFile;
   }

   public void setConfigSpec(String userCfgSpec) throws Exception {
      if (userCfgSpec!=null) {
         parseConfigFileNames(userCfgSpec);         
      }
   }

   /**
    * Calculate the maximum number of servers in a given policy file
    * Output: integer value of max array length
    */
   public int getRuleMaxLen(Hashtable<String, String[] > aTable) {
      int len;
      int longlen=0;
      String key;
      Vector<Integer> v = new Vector<Integer>();

      // Iterate through master table to get key lengths 
      Iterator<String> itString = aTable.keySet().iterator(); // Iterator of all keys
      while (itString.hasNext()){
         key = itString.next();
         // don't count length of req validation rules
         if (key.equalsIgnoreCase(HAUtil.TRANSID) | key.equalsIgnoreCase(HAUtil.USERID) | key.equalsIgnoreCase(HAUtil.PROGID) ){
            len = 0;
         } else {
            len = aTable.get(key).length;
            v.add(len);
            this.debugln("Server rule (%s) length: %d",key,len);
         }        
      }

      // Iterate through length vector to get key lengths
      Iterator<Integer> itInt = v.iterator();
      while (itInt.hasNext()){
         len = itInt.next();
         if (len>longlen){
            longlen=len;
         }     
      }     
      return longlen;
   }

}
