/*
 *  Licensed Materials - Property of IBM
 *  
 *  CICS SupportPac CA1T
 * 
 *   Copyright IBM Corporation 2008, 2011
 *  All Rights Reserved
 *  
 *  US Government Users Restricted Rights  Use,
 *  duplication, or disclosure restricted by GSA ADP
 *  Schedule Contract with IBM Corporation
 */

/* 
 *      File Name:    RoundRobin.java
 *      Version:      CA1T V2.1 
 *      Date:         26042011
 *      Product:      CICS Transaction Gateway for z/OS
 * 
 * Description: The RoundRobin exit is designed to redirect an ECI request across
 * a group of cloned CICS servers
 * 
 * Change History
 * V1.0 - GA version
 * V1.1 - Comment parsing
 * V2.0 - Retry interval and V8 LastServers support
 * V2.1 - Add support for an alternate policy   
 * 
 * See SupportPac documentation for further details on how to configure and use
 * this sample
 */


package com.ibm.ctg.samples.ha;

import com.ibm.ctg.ha.CICSRequestExit;
import com.ibm.ctg.ha.RequestDetails;
import java.util.*;

public class RoundRobinExit extends HAExit implements CICSRequestExit {
	//                                             
	public static final String copyright = "\nLicensed Materials - Property of IBM "+
	"CICS SupportPac CA1T" +
	"(c) Copyright IBM Corp. 2008, 2011  All Rights Reserved. " +
	"US Government Users Restricted Rights - Use, duplication or " +
	"disclosure restricted by GSA ADP Schedule Contract with " +
	"IBM Corp.\n";
	//
	   
    /**
     * Constructor called at exit initialisation by Gateway 
     */
    public RoundRobinExit() throws Exception {
    		
    	super(); // Invoke super class constructor			
   
    }
    
	/**
	 * Maps a server alias to a server
	 *  
	 * Input: alias, current time, request map 
	 * Returns: server to be used
	 */
	public synchronized String mapServer(String alias, Date d, Map<RequestDetails, Object> data ) throws NoCICSServersException {
		String[] array = configTable.get(alias); // Array of possible servers
		String server = array[0];// Initialise server to fist array member in case all are dirty
		int serverIndex = 0;
		int retry = (Integer) data.get(RequestDetails.RetryCount);
		boolean foundClean=false;
		util.debugln("RoundRobinExit.mapServer(alias=%s,date=%s,retries=%d)",alias,d,retry);
		
		// if first time - index is what we set it to last time
		if (retry == 0) {
			serverIndex = serverTable.get(alias); // get the serverindex
		} else {  	
	      // else retrying - so index is based on the last server used
			String lastserver = data.get(RequestDetails.LastServer).toString();
			
			util.debugln("Locating array index of previous failing CICS server(%s)",lastserver);
			for (int x = 0; x < array.length; x++) { // find the index in the array													
				if (array[x].equals(lastserver)) {
					serverIndex =(x + 1) % array.length ; //move index on by one
					break;
				}
			}
		}
		util.debugln("Index = %d",serverIndex);
	
		// Now loop around the array until we find a clean server
		for (int x = 0; x < array.length; x++ ) {
			server = array[serverIndex]; // get the server name for this index
			if (isClean(server,d)) {							
            foundClean=true;
            util.debugln("CICS server %s is clean (available)",server);
				break; // exit for loop with server name
         } else {
            util.debugln("CICS server %s is dirty (unavailable)",server);
            // Server is dirty so lets go round again
         }
			serverIndex = (serverIndex + 1) % array.length; // increment the array index and check again						
		} 
		nextIndex = (serverIndex + 1) % array.length;  // set the index if we retry
      serverTable.put(alias, nextIndex); // store the nextindex in the server
                                         // table
      
      if (!foundClean) {
         // Didn't find any available servers under the current policy
         // Throw NoCICSServersException
         throw new NoCICSServersException("RoundRobinExit.mapServer() Last CICS server tried="+server);
      }
	
		return server;
	}

}

